import os
import signal

import pytest

import signalfd

def test_block_and_read():
    fd = signalfd.signalfd(-1, [signal.SIGUSR1], signalfd.SFD_CLOEXEC)
    try:
        signalfd.sigprocmask(signalfd.SIG_BLOCK, [signal.SIGUSR1])
        os.kill(os.getpid(), signal.SIGUSR1)
        si = signalfd.read_siginfo(fd)
        assert si.ssi_signo == signal.SIGUSR1
    finally:
        os.close(fd)


def test_read_closed():
    fd = signalfd.signalfd(-1, [signal.SIGUSR1], signalfd.SFD_CLOEXEC)
    os.close(fd)
    pytest.raises(OSError, signalfd.read_siginfo, fd)



from cffi import FFI

ffi = FFI()
ffi.cdef("""
#define SFD_CLOEXEC ...
#define SFD_NONBLOCK ...
struct signalfd_siginfo {
    uint32_t ssi_signo;   /* Signal number */
    int32_t  ssi_errno;   /* Error number (unused) */
    int32_t  ssi_code;    /* Signal code */
    uint32_t ssi_pid;     /* PID of sender */
    uint32_t ssi_uid;     /* Real UID of sender */
    int32_t  ssi_fd;      /* File descriptor (SIGIO) */
    uint32_t ssi_tid;     /* Kernel timer ID (POSIX timers)
    uint32_t ssi_band;    /* Band event (SIGIO) */
    uint32_t ssi_overrun; /* POSIX timer overrun count */
    uint32_t ssi_trapno;  /* Trap number that caused signal */
    int32_t  ssi_status;  /* Exit status or signal (SIGCHLD) */
    int32_t  ssi_int;     /* Integer sent by sigqueue(3) */
    uint64_t ssi_ptr;     /* Pointer sent by sigqueue(3) */
    uint64_t ssi_utime;   /* User CPU time consumed (SIGCHLD) */
    uint64_t ssi_stime;   /* System CPU time consumed (SIGCHLD) */
    uint64_t ssi_addr;    /* Address that generated signal (for hardware-generated signals) */
    ...; // uint8_t  __pad[48];
                          /* Pad size to 128 bytes (allow for additional fields in the future) */
};
""")

ffi.cdef("""
typedef struct {
    unsigned long int __val[%d];
} sigset_t;
int signalfd(int fd, const sigset_t *mask, int flags);
""" % (1024 / 8 * ffi.sizeof('unsigned long int')))

ffi.cdef("""
int sigemptyset(sigset_t *set);
int sigfillset(sigset_t *set);
int sigaddset(sigset_t *set, int signum);
int sigdelset(sigset_t *set, int signum);
int sigismember(const sigset_t *set, int signum);
""")

ffi.cdef("""
#define SIG_BLOCK ...
#define SIG_UNBLOCK ...
#define SIG_SETMASK ...
int pthread_sigmask(int how, const sigset_t *set, sigset_t *oldset);
int sigprocmask(int how, const sigset_t *set, sigset_t *oldset);
""")

ffi.set_source("signalfd._signalfd", """
#include <sys/signalfd.h>
#include <stdint.h> /* Definition of uint64_t */
#include <signal.h>
""")

if __name__ == "__main__":
    ffi.compile()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import, print_function

import io
import re
import sys
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    return io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ).read()

setup(
    name='signalfd',
    version='0.1.0',
    license='BSD',
    description='CFFI bindings for signalfd.',
    long_description='%s\n%s' % (read('README.rst'), re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))),
    author='Ionel Cristian Mărieș',
    author_email='contact@ionelmc.ro',
    url='https://github.com/ionelmc/python-signalfd',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities',
    ],
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    setup_requires=[
        "cffi>=1.0.0"
    ] if any(i.startswith('build') or i.startswith('bdist') for i in sys.argv) else [],
    install_requires=[
        "cffi>=1.0.0",
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    cffi_modules=[i + ":ffi" for i in glob("src/*/_*_build.py")],
)



"""
Simple wrapper around the signalfd syscall.
"""
import errno
import os

from ._signalfd import ffi as _ffi
from ._signalfd import lib as _lib

__version__ = "0.1.0"

SFD_CLOEXEC = _lib.SFD_CLOEXEC
SFD_NONBLOCK = _lib.SFD_NONBLOCK

SIG_BLOCK = _lib.SIG_BLOCK
SIG_UNBLOCK = _lib.SIG_UNBLOCK
SIG_SETMASK = _lib.SIG_SETMASK

SIGINFO_SIZE = _ffi.sizeof('struct signalfd_siginfo')


class UnknownError(Exception):
    pass


def signalfd(fd, signals, flags):
    if hasattr(fd, 'fileno'):
        fd = fd.fileno()

    if not isinstance(fd, int):
        raise TypeError('fd: must be an integer.')
    if not isinstance(flags, int):
        raise TypeError('flags: must be an integer.')

    sigmask = _ffi.new('sigset_t[1]')
    for sig in signals:
        _lib.sigaddset(sigmask, sig)

    result = _lib.signalfd(fd, sigmask, flags)

    if result < 0:
        err = _lib.errno
        if err == errno.EBADF:
            raise ValueError("fd: not a valid file descriptor.")
        elif err == errno.EINVAL:
            if flags & (0xffffffff ^ (SFD_CLOEXEC | SFD_NONBLOCK)):
                raise ValueError("flags: mask contains invalid values.")
            else:
                raise ValueError("fd: not a signalfd.")
        elif err == errno.EMFILE:
            raise OSError("max system fd limit reached.")
        elif err == errno.ENFILE:
            raise OSError("max system fd limit reached.")
        elif err == errno.ENODEV:
            raise OSError("could not mount (internal) anonymous inode device.")
        elif err == errno.ENOMEM:
            raise MemoryError("insufficient kernel memory available.")
        else:
            raise UnknownError(err)

    return result


def sigprocmask(flags, signals):
    if not isinstance(flags, int):
        raise TypeError('flags: must be an integer.')

    new_sigmask = _ffi.new('sigset_t[1]')
    old_sigmask = _ffi.new('sigset_t[1]')

    _lib.sigemptyset(new_sigmask)
    for sig in signals:
        _lib.sigaddset(new_sigmask, sig)

    ret = _lib.pthread_sigmask(flags, new_sigmask, old_sigmask)

    if ret < 0:
        err = _ffi.errno
        if err == errno.EINVAL:
            raise ValueError("flags: invalid value (not one of SIG_BLOCK, SIG_UNBLOCK or SIG_SETMASK)")
        elif err == errno.EFAULT:
            raise ValueError("sigmask is not a valid sigset_t")
        else:
            raise UnknownError(err)


def read_siginfo(fh):
    info = _ffi.new('struct signalfd_siginfo *')
    buffer = _ffi.buffer(info)

    if hasattr(fh, 'readinto'):
        fh.readinto(buffer)
    else:
        buffer[:] = os.read(fh, SIGINFO_SIZE)
    return info
